from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Literal, NotRequired, Required, Self, TypedDict, Unpack

from gatelogue_types._util import _Column, _FKColumn, _format_code, _format_str, _SetAttr
from gatelogue_types.node import LocatedNode, Node

if TYPE_CHECKING:
    import sqlite3
    from collections.abc import Iterator

type SeaMode = Literal["cruise", "warp ferry", "traincarts ferry"]


class SeaCompany(Node):
    name = _Column[str]("name", "SeaCompany", formatter=_format_str)
    """Name of the sea company"""
    link = _Column[str | None]("link", "SeaCompany", sourced=True, formatter=_format_str)
    """Link to the MRT Wiki page for the company"""
    COLUMNS: ClassVar = (name, link)

    def __str__(self):
        return super().__str__() + f" {self.name}"

    class CreateParams(TypedDict, total=False):
        """Internal use"""

        name: Required[str]
        link: str | None

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node(conn, src, ty=cls.__name__)
        cur = conn.cursor()
        cur.execute("INSERT INTO SeaCompany (i, name, link) VALUES (:i, :name, :link)", dict(i=i, **kwargs))
        cur.execute("INSERT INTO SeaCompanySource (i, source) VALUES (:i, :source)", dict(i=i, source=src, **kwargs))
        return cls(conn, i)

    @property
    def lines(self) -> Iterator[SeaLine]:
        """List of all :py:class:`SeaLine` s the company operates"""
        return (
            SeaLine(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM SeaLine WHERE company = :i", dict(i=self.i)).fetchall()
        )

    @property
    def stops(self) -> Iterator[SeaStop]:
        """List of all :py:class:`SeaStop` s the company's lines stop at"""
        return (
            SeaStop(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM SeaStop WHERE company = :i", dict(i=self.i)).fetchall()
        )

    @property
    def docks(self) -> Iterator[SeaDock]:
        """List of all :py:class:`SeaDock` s the company's lines stop at"""
        return (
            SeaDock(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaDock.i "
                "FROM (SELECT i FROM SeaStop WHERE company = :i) A "
                "INNER JOIN SeaDock on A.i = SeaDock.stop",
                dict(i=self.i),
            ).fetchall()
        )

    def equivalent_nodes(self) -> Iterator[Self]:
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM SeaCompany WHERE name = ?", (self.name,)).fetchall()
        )

    def _merge(self, other: Self):
        cur = self.conn.cursor()
        cur.execute("UPDATE SeaLine SET company = :i1 WHERE company = :i2", dict(i1=self.i, i2=other.i))
        cur.execute("UPDATE SeaStop SET company = :i1 WHERE company = :i2", dict(i1=self.i, i2=other.i))


class SeaLine(Node):
    code = _Column[str]("code", "SeaLine", formatter=_format_str)
    """Unique code identifying the sea line"""
    company = _FKColumn(SeaCompany, "company", "SeaLine")
    """The :py:class:`SeaCompany` that operates the line"""
    name = _Column[str | None]("name", "SeaLine", sourced=True, formatter=_format_str)
    """Name of the line"""
    colour = _Column[str | None]("colour", "SeaLine", sourced=True, formatter=_format_str)
    """Colour of the line (on a map)"""
    mode = _Column[SeaMode | None]("mode", "SeaLine", sourced=True, formatter=_format_str)
    """Type of sea vehicle or technology the line uses"""
    local = _Column[bool | None]("local", "SeaLine", sourced=True)
    """Whether the company operates within the city, e.g. a local ferry service"""
    COLUMNS: ClassVar = (code, company, name, colour, mode, local)

    def __str__(self):
        return super().__str__() + f" {self.company.name} {self.code}"

    class CreateParams(TypedDict, total=False):
        """Internal use"""

        code: Required[str]
        company: Required[SeaCompany]
        name: str | None
        colour: str | None
        mode: SeaMode | None
        local: bool | None

    @classmethod
    def create(
        cls,
        conn: sqlite3.Connection,
        src: int,
        **kwargs: Unpack[CreateParams],
    ) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node(conn, src, ty=cls.__name__)
        cur = conn.cursor()
        cur.execute(
            "INSERT INTO SeaLine (i, code, company, name, colour, mode, local) "
            "VALUES (:i, :code, :company, :name, :colour, :mode, :local)",
            dict(i=i, **kwargs),
        )
        cur.execute(
            "INSERT INTO SeaLineSource (i, source, name, colour, mode, local) "
            "VALUES (:i, :source, :name_src, :colour_src, :mode_src, :local_src)",
            dict(i=i, source=src, **kwargs),
        )
        return cls(conn, i)

    @property
    def docks(self) -> Iterator[SeaDock]:
        """List of all :py:class:`SeaDocks` s the line stops at"""
        return (
            SeaDock(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaDock.i "
                'FROM (SELECT "from", "to" FROM SeaConnection WHERE line = :i) A '
                'LEFT JOIN SeaDock ON A."from" = SeaDock.i OR A."to" = SeaDock.i',
                dict(i=self.i),
            ).fetchall()
        )

    @property
    def stops(self) -> Iterator[SeaStop]:
        """List of all :py:class:`SeaStop` s the line stops at"""
        return (
            SeaStop(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaDock.stop "
                'FROM (SELECT "from", "to" FROM SeaConnection WHERE line = :i) A '
                'LEFT JOIN SeaDock ON A."from" = SeaDock.i OR A."to" = SeaDock.i',
                dict(i=self.i),
            ).fetchall()
        )

    def equivalent_nodes(self) -> Iterator[Self]:
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT i FROM SeaLine WHERE code = ? AND company = ?", (self.code, self.company.i)
            ).fetchall()
        )

    def _merge(self, other: Self):
        self.conn.execute("UPDATE SeaConnection SET line = :i1 WHERE line = :i2", dict(i1=self.i, i2=other.i))


class SeaStop(LocatedNode):
    codes = _SetAttr[str]("SeaStopCodes", "code", formatter=_format_code)
    """Unique code(s) identifying the sea stop. May also be the same as the name"""
    company = _FKColumn(SeaCompany, "company", "SeaStop")
    """The :py:class:`SeaCompany` that owns this stop"""
    name = _Column[str | None]("name", "SeaStop", sourced=True, formatter=_format_str)
    """Name of the stop"""
    COLUMNS: ClassVar = (*LocatedNode.COLUMNS, codes, company, name)

    def __str__(self):
        return super().__str__() + f" {self.company.name} {'/'.join(self.codes)}"

    class CreateParams(LocatedNode.CreateParams, total=False):
        codes: Required[set[str]]
        company: Required[SeaCompany]
        name: str | None

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        codes = kwargs.get("codes", set())
        i = cls.create_node_with_location(conn, src, ty=cls.__name__, **kwargs)
        cur = conn.cursor()
        cur.execute(
            "INSERT INTO SeaStop (i, name, company) VALUES (:i, :name, :company)",
            dict(i=i, **kwargs),
        )
        cur.executemany("INSERT INTO SeaStopCodes (i, code) VALUES (?, ?)", [(i, code) for code in codes])
        cur.execute(
            "INSERT INTO SeaStopSource (i, source, name) VALUES (:i, :source, :name_src)",
            dict(i=i, source=src, **kwargs),
        )
        return cls(conn, i)

    @property
    def docks(self) -> Iterator[SeaDock]:
        """List of :py:class:`SeaDock` s this stop has"""
        return (
            SeaDock(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT i FROM SeaDock WHERE stop = :i",
                dict(i=self.i),
            ).fetchall()
        )

    @property
    def connections_from_here(self) -> Iterator[SeaConnection]:
        """List of all :py:class:`SeaConnections` s departing from this stop"""
        return (
            SeaConnection(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaConnection.i "
                "FROM (SELECT i FROM SeaDock WHERE stop = :i) A "
                'INNER JOIN SeaConnection ON A.i = SeaConnection."from"',
                dict(i=self.i),
            ).fetchall()
        )

    @property
    def connections_to_here(self) -> Iterator[SeaConnection]:
        """List of all :py:class:`SeaConnections` s arriving at this stop"""
        return (
            SeaConnection(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaConnection.i "
                "FROM (SELECT i FROM SeaDock WHERE stop = :i) A "
                'INNER JOIN SeaConnection ON A.i = SeaConnection."to"',
                dict(i=self.i),
            ).fetchall()
        )

    @property
    def lines(self) -> Iterator[SeaLine]:
        """List of all :py:class:`SeaLines` s at this stop"""
        return (
            SeaLine(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaConnection.line "
                "FROM (SELECT i FROM SeaDock WHERE stop = :i) A "
                'LEFT JOIN SeaConnection ON A.i = SeaConnection."from" OR A.i = SeaConnection."to"',
                dict(i=self.i),
            ).fetchall()
        )

    def equivalent_nodes(self) -> Iterator[Self]:
        if len(codes := self.codes) == 0:
            return iter(())
        code_params = "?, " * (len(codes) - 1) + "?"
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaStop.i "
                "FROM SeaStop LEFT JOIN SeaStopCodes ON SeaStop.i = SeaStopCodes.i "
                f"WHERE company = ? AND SeaStopCodes.code IN ({code_params})",
                (
                    self.company.i,
                    *codes,
                ),
            ).fetchall()
        )

    def _merge(self, other: Self):
        cur = self.conn.cursor()
        cur.execute("UPDATE SeaDock SET stop = :i1 WHERE stop = :i2", dict(i1=self.i, i2=other.i))
        super()._merge(other)


class SeaDock(Node):
    code = _Column[str | None]("code", "SeaDock", formatter=_format_code)
    """Unique code identifying the dock. May not necessarily be the same as the code ingame. If ``None``, code is unspecified"""
    stop = _FKColumn(SeaStop, "stop", "SeaDock")
    """The :py:class:`SeaStop` of the dock"""
    COLUMNS: ClassVar = (code, stop)

    def __str__(self):
        return super().__str__() + f" {self.stop.company.name} {'/'.join(self.stop.codes)} {self.code}"

    class CreateParams(TypedDict):
        code: str | None
        stop: SeaStop

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node(conn, src, ty=cls.__name__)
        cur = conn.cursor()
        cur.execute("INSERT INTO SeaDock (i, code, stop) VALUES (:i, :code, :stop)", dict(i=i, **kwargs))
        cur.execute("INSERT INTO SeaDockSource (i, source) VALUES (:i, :source)", dict(i=i, source=src, **kwargs))
        return cls(conn, i)

    @property
    def connections_from_here(self) -> Iterator[SeaConnection]:
        """List of all :py:class:`SeaConnection` s departing from this dock"""
        return (
            SeaConnection(self.conn, i)
            for (i,) in self.conn.execute(
                'SELECT SeaConnection.i FROM SeaConnection WHERE SeaConnection."from" = :i',
                dict(i=self.i),
            ).fetchall()
        )

    @property
    def connections_to_here(self) -> Iterator[SeaConnection]:
        """List of all :py:class:`SeaConnection` s arriving at this dock"""
        return (
            SeaConnection(self.conn, i)
            for (i,) in self.conn.execute(
                'SELECT SeaConnection.i FROM SeaConnection WHERE SeaConnection."to" = :i',
                dict(i=self.i),
            ).fetchall()
        )

    @property
    def lines(self) -> Iterator[SeaLine]:
        """List of all :py:class:`SeaLine` s at this dock"""
        return (
            SeaLine(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT SeaConnection.line FROM SeaConnection "
                'WHERE SeaConnection."from" = :i OR SeaConnection."to" = :i',
                dict(i=self.i),
            ).fetchall()
        )

    def equivalent_nodes(self) -> Iterator[Self]:
        if (code := self.code) is None:
            return iter(())
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT i FROM SeaDock WHERE stop = ? AND code = ?", (self.stop.i, code)
            ).fetchall()
        )

    def _merge(self, other: Self):
        cur = self.conn.cursor()
        cur.execute('UPDATE SeaConnection SET "from" = :i1 WHERE "from" = :i2', dict(i1=self.i, i2=other.i))
        cur.execute('UPDATE SeaConnection SET "to" = :i1 WHERE "to" = :i2', dict(i1=self.i, i2=other.i))


class SeaConnection(Node):
    line = _FKColumn(SeaLine, "line", "SeaConnection")
    """The :py:class:`SeaLine` that the connection is made on"""
    from_ = _FKColumn(SeaDock, "from", "SeaConnection")
    """The :py:class:`SeaDock` the connection departs from"""
    to = _FKColumn(SeaDock, "to", "SeaConnection")
    """The :py:class:`SeaDock` the connection arrives at"""
    direction = _Column[str | None]("direction", "SeaConnection", sourced=True, formatter=_format_str)
    """The direction taken when travelling along this connection, e.g. ``Eastbound``, ``towards Terminus Name``"""
    duration = _Column[int | None]("duration", "SeaConnection", sourced=True, formatter=_format_str)
    """The duration taken in seconds to travel on this connection"""
    COLUMNS: ClassVar = (line, from_, to, direction, duration)

    def __str__(self):
        return super().__str__() + f" {self.line.company.name} {self.line.code} {self.from_.code} -> {self.to.code}"

    class CreateParams(TypedDict):
        line: SeaLine
        from_: SeaDock
        to: SeaDock
        direction: NotRequired[str | None]
        duration: NotRequired[int | None]

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node(conn, src, ty=cls.__name__)
        cur = conn.cursor()
        cur.execute(
            'INSERT INTO SeaConnection (i, line, "from", "to", direction, duration) VALUES (:i, :line, :from_, :to, :direction, :duration)',
            dict(i=i, **kwargs),
        )
        cur.execute(
            "INSERT INTO SeaConnectionSource (i, source, direction, duration) VALUES (:i, :source, :direction_src, :duration_src)",
            dict(i=i, source=src, **kwargs),
        )
        return cls(conn, i)

    @classmethod
    def create2(
        cls, conn: sqlite3.Connection, src: int, *, direction2: str | None = None, **kwargs: Unpack[CreateParams]
    ) -> tuple[Self, Self]:
        kwargs2 = kwargs.copy()
        kwargs2["from_"], kwargs2["to"] = kwargs2["to"], kwargs2["from"]
        kwargs2["direction"] = direction2 or kwargs2["direction"]
        return cls.create(conn, src, **kwargs), cls.create(conn, src, **kwargs2)

    def equivalent_nodes(self) -> Iterator[Self]:
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                'SELECT i FROM SeaConnection WHERE line = ? AND "from" = ? AND "to" = ?',
                (self.line.i, self.from_.i, self.to.i),
            ).fetchall()
        )



from gatelogue_types import GD
from gatelogue_types.air import AirAirline, AirAirport, AirFlight, AirGate
from gatelogue_types.bus import BusCompany, BusStop


def test_get():
    gd = GD.urllib_get(sources=True)
    print(gd.timestamp, gd.version)
    assert gd.has_sources
    assert True


def test_get_no_sources():
    gd = GD.urllib_get()
    print(gd.timestamp, gd.version)
    assert not gd.has_sources
    assert True


def test_air_gate_merge():
    gd = GD.create(["0"])

    airport = AirAirport.create(gd.conn, 0, code="AAA")
    gate1 = AirGate.create(gd.conn, 0, airport=airport, code=None)
    gate2 = AirGate.create(gd.conn, 0, airport=airport, code="1", size="S")
    assert gd.conn.execute("SELECT count(rowid) FROM AirGate").fetchone()[0] == 2
    gate1.merge(gate2)
    assert gd.conn.execute("SELECT count(rowid) FROM AirGate").fetchone()[0] == 1


def test_air_flight_merge():
    gd = GD.create(["0", "1"])

    airline = AirAirline.create(gd.conn, 0, name="Example Air")
    airport1 = AirAirport.create(gd.conn, 0, code="AAA")
    airport2 = AirAirport.create(gd.conn, 0, code="BBB")
    gate1 = AirGate.create(gd.conn, 0, airport=airport1, code=None, mode="helicopter")
    gate2 = AirGate.create(gd.conn, 0, airport=airport2, code=None, mode="helicopter")
    gate3 = AirGate.create(gd.conn, 1, airport=airport1, code="A")
    gate4 = AirGate.create(gd.conn, 1, airport=airport2, code="B")
    flight1 = AirFlight.create(gd.conn, 0, airline=airline, code="001", from_=gate1, to=gate2)
    flight2 = AirFlight.create(gd.conn, 0, airline=airline, code="001", from_=gate3, to=gate4)
    assert gd.conn.execute("SELECT count(rowid) FROM AirFlight").fetchone()[0] == 2
    assert gd.conn.execute("SELECT count(rowid) FROM AirGate").fetchone()[0] == 4
    assert flight2 in flight1.equivalent_nodes()
    flight1.merge(flight2)
    assert gd.conn.execute("SELECT count(rowid) FROM AirFlight").fetchone()[0] == 1
    assert gd.conn.execute("SELECT count(rowid) FROM AirGate").fetchone()[0] == 2
    assert flight1.from_.code == "A"
    assert flight1.to.code == "B"
    assert flight1.from_.mode == "helicopter"
    assert flight1.to.mode == "helicopter"


def test_bus_stop_merge():
    gd = GD.create(["0", "1"])

    company = BusCompany.create(gd.conn, 0, name="Example Inc")
    stop1 = BusStop.create(gd.conn, 0, codes={"a", "b"}, company=company)
    stop2 = BusStop.create(gd.conn, 1, codes={"b", "c"}, company=company)
    assert stop2 in stop1.equivalent_nodes()
    stop1.merge(stop2)



from __future__ import annotations

import re
import warnings
from typing import TYPE_CHECKING, Literal, LiteralString

if TYPE_CHECKING:
    from collections.abc import Callable, Iterator

    from gatelogue_types import Aircraft
    from gatelogue_types.node import Node


def _format_str[T: str](s: T | None) -> T | None:
    if s is None or s.strip().lower() in ("", "?", "??", "-"):
        return None
    return s.strip()


def _format_code[T: str](s: T | None) -> T | None:
    def search_all(regex: re.Pattern[str], text: str) -> Iterator[re.Match[str]]:
        pos = 0
        while (m := regex.search(text, pos)) is not None:
            pos = m.end()
            yield m

    if (s := _format_str(s)) is None:
        return None
    res = ""
    hyphen1 = False
    hyphen2 = False
    for match in search_all(re.compile(r"\d+|[A-Za-z]+|[^\dA-Za-z]+"), str(s).strip()):
        t = match.group(0)
        if len(t) == 0:
            continue
        if (
            (hyphen1 and t[0].isdigit())
            or (hyphen2 and t[0].isalpha())
            or (len(res) != 0 and t[0].isdigit() and res[-1].isdigit())
        ):
            res += "-"
        if hyphen1:
            hyphen1 = False
        if hyphen2:
            hyphen2 = False
        if t.isdigit():
            res += t.lstrip("0") or "0"
        elif t.isalpha():
            res += t.upper()
        elif (t == "-" and len(res) == 0) or (len(res) != 0 and res[-1].isdigit()):
            hyphen1 = True
        elif len(res) != 0 and res[-1].isalpha():
            hyphen2 = True

    return res


def _warn_clash(
    warn_fn: Callable[[str], object],
    column: str,
    table: str,
    str_instance1: str,
    self_v,
    str_instance2: str,
    other_v,
    self_sources: set[int],
    other_sources: set[int],
    priority: Literal["former", "latter"],
):
    warn_fn(
        f"{column} in table {table} is different "
        f"between {str_instance1} ({self_v}) and {str_instance2} ({other_v}). "
        + (
            f"Former has higher priority of {self_sources} than latter which has {other_sources}"
            if priority == "former"
            else f"Latter has higher priority of {other_sources} than former which has {self_sources}"
        )
    )


class _Column[T]:
    def __init__(
        self,
        name: LiteralString,
        table: LiteralString,
        sourced: bool = False,
        formatter: Callable[[T], T] | None = None,
    ):
        self.name = f'"{name}"'
        self.table = table
        self.sourced = sourced
        self.formatter = formatter

    def __get__(self, instance: Node, owner: type[Node]) -> T:
        return instance.conn.execute(
            f"SELECT {self.name} FROM {self.table} WHERE i = :i", dict(i=instance.i)
        ).fetchone()[0]

    def __set__(self, instance: Node, value: T | tuple[set[int], T]):
        srcs, value = value if isinstance(value, tuple) else (instance.sources, value)
        if self.formatter is not None:
            value = self.formatter(value)
        cur = instance.conn.cursor()
        old_value = self.__get__(instance, type(instance)) if self.sourced else None
        cur.execute(f"UPDATE {self.table} SET {self.name} = :value WHERE i = :i", dict(value=value, i=instance.i))
        if not self.sourced:
            return
        if value != old_value:
            cur.execute(f"UPDATE {self.table + 'Source'} SET {self.name} = false WHERE i = :i", dict(i=instance.i))
        for src in srcs:
            cur.execute(
                f"INSERT INTO {self.table + 'Source'} (i, source, {self.name}) VALUES (:i, :source, :bool) "
                f"ON CONFLICT (i, source) DO UPDATE SET {self.name} = :bool",
                dict(bool=value is not None, i=instance.i, source=src),
            )

    def sources(self, instance: Node) -> set[int]:
        return {
            src
            for (src,) in instance.conn.execute(
                f"SELECT DISTINCT source FROM {self.table + 'Source'} WHERE i = :i AND {self.name} = true",
                dict(i=instance.i),
            ).fetchall()
        }

    def _merge(
        self,
        instance1: Node,
        instance2: Node,
        str_instance1: str | None = None,
        str_instance2: str | None = None,
        warn_fn: Callable[[str], object] = warnings.warn,
    ):
        str_instance1 = str_instance1 or str(instance1)
        str_instance2 = str_instance2 or str(instance2)
        self_v = self.__get__(instance1, type(instance1))
        other_v = self.__get__(instance2, type(instance2))
        if not self.sourced:
            if self_v != other_v:
                self_sources = instance1.sources
                other_sources = instance2.sources
                if min(self_sources) < min(other_sources):
                    if self_sources != other_sources:
                        _warn_clash(
                            warn_fn,
                            "Column " + self.name,
                            self.table,
                            str_instance1,
                            self_v,
                            str_instance2,
                            other_v,
                            self_sources,
                            other_sources,
                            "former",
                        )
                else:
                    if self_sources != other_sources:
                        _warn_clash(
                            warn_fn,
                            "Column " + self.name,
                            self.table,
                            str_instance1,
                            self_v,
                            str_instance2,
                            other_v,
                            self_sources,
                            other_sources,
                            "latter",
                        )
                    self.__set__(instance1, (self_sources, other_v))
            return
        match (self_v, other_v):
            case (None, None):
                pass
            case (_, None):
                pass
            case (None, _):
                sources = self.sources(instance2)
                self.__set__(instance1, (sources, other_v))
            case (_, _) if self_v != other_v:
                self_sources = self.sources(instance1)
                other_sources = self.sources(instance2)
                if min(self_sources) < min(other_sources):
                    if self_sources != other_sources:
                        _warn_clash(
                            warn_fn,
                            "Column " + self.name,
                            self.table,
                            str_instance1,
                            self_v,
                            str_instance2,
                            other_v,
                            self_sources,
                            other_sources,
                            "former",
                        )
                    self.__set__(instance2, None)
                else:
                    if self_sources != other_sources:
                        _warn_clash(
                            warn_fn,
                            "Column " + self.name,
                            self.table,
                            str_instance1,
                            self_v,
                            str_instance2,
                            other_v,
                            self_sources,
                            other_sources,
                            "latter",
                        )
                    self.__set__(instance1, (other_sources, other_v))


class _CoordinatesColumn:
    name = "coordinates"

    def __get__(self, instance: Node, owner: type[Node]) -> tuple[int, int] | None:
        x, y = instance.conn.execute("SELECT x, y from NodeLocation WHERE i = :i", dict(i=instance.i)).fetchone()
        return None if x is None or y is None else (x, y)

    def __set__(self, instance: Node, value: tuple[int, int] | None | tuple[set[int], tuple[int, int] | None]):
        srcs, value = value if value is not None and isinstance(value[0], set) else (instance.sources, value)
        x, y = (None, None) if value is None else value
        cur = instance.conn.cursor()
        old_value = self.__get__(instance, type(instance))
        cur.execute("UPDATE NodeLocation SET x = :x, y = :y WHERE i = :i", dict(x=x, y=y, i=instance.i))

        if value != old_value:
            cur.execute("UPDATE NodeLocationSource SET coordinates = false WHERE i = :i", dict(i=instance.i))
        for src in srcs:
            cur.execute(
                "INSERT INTO NodeLocationSource (i, source, coordinates) VALUES (:i, :source, :bool) "
                "ON CONFLICT (i, source) DO UPDATE SET coordinates = :bool",
                dict(bool=value is not None, i=instance.i, source=src),
            )

    @staticmethod
    def sources(instance: Node) -> set[int]:
        return {
            src
            for (src,) in instance.conn.execute(
                "SELECT DISTINCT source FROM NodeLocationSource WHERE i = :i AND coordinates = true", dict(i=instance.i)
            ).fetchall()
        }

    def _merge(
        self,
        instance1: Node,
        instance2: Node,
        str_instance1: str | None = None,
        str_instance2: str | None = None,
        warn_fn: Callable[[str], object] = warnings.warn,
    ):
        str_instance1 = str_instance1 or str(instance1)
        str_instance2 = str_instance2 or str(instance2)
        self_v = self.__get__(instance1, type(instance1))
        other_v = self.__get__(instance2, type(instance2))
        match (self_v, other_v):
            case (None, None):
                pass
            case (_, None):
                pass
            case (None, _):
                sources = self.sources(instance2)
                self.__set__(instance1, (sources, other_v))
            case (_, _) if self_v != other_v:
                self_sources = self.sources(instance1)
                other_sources = self.sources(instance2)
                if min(self_sources) < min(other_sources):
                    if self_sources != other_sources:
                        _warn_clash(
                            warn_fn,
                            "Columns x/y",
                            "NodeLocation",
                            str_instance1,
                            self_v,
                            str_instance2,
                            other_v,
                            self_sources,
                            other_sources,
                            "former",
                        )
                    self.__set__(instance2, None)
                else:
                    if self_sources != other_sources:
                        _warn_clash(
                            warn_fn,
                            "Columns x/y",
                            "NodeLocation",
                            str_instance1,
                            self_v,
                            str_instance2,
                            other_v,
                            self_sources,
                            other_sources,
                            "latter",
                        )
                    self.__set__(instance1, (other_sources, other_v))


class _FKColumn[T: Node | None]:
    def __init__(self, ty: type[T], name: LiteralString, table: LiteralString, sourced: bool = False):
        self.name = name
        self.table = table
        self.sourced = sourced
        self.ty = ty

    def __get__(self, instance: Node, owner: type[Node]) -> T:
        target_i = _Column(self.name, self.table, self.sourced).__get__(instance, owner)
        if target_i is None:
            return None
        return self.ty(instance.conn, target_i)

    def __set__(self, instance: Node, value: T | tuple[int, T]):
        _Column(self.name, self.table, self.sourced).__set__(instance, None if value is None else value.i)

    def _merge(
        self,
        instance1: Node,
        instance2: Node,
        str_instance1: str | None = None,
        str_instance2: str | None = None,
        warn_fn: Callable[[str], object] = warnings.warn,
    ):
        _Column(self.name, self.table, self.sourced)._merge(instance1, instance2, str_instance1, str_instance2, warn_fn)


class _AircraftColumn:
    def __init__(self, name: LiteralString, table: LiteralString, sourced: bool = False):
        self.name = name
        self.table = table
        self.sourced = sourced

    def __get__(self, instance: Node, owner: type[Node]) -> Aircraft | None:
        target_name = _Column(self.name, self.table, self.sourced).__get__(instance, owner)
        if target_name is None:
            return None
        from gatelogue_types import Aircraft

        return Aircraft(instance.conn, target_name)

    def __set__(self, instance: Node, value: Aircraft | tuple[int, Aircraft]):
        _Column(self.name, self.table, self.sourced).__set__(instance, None if value is None else value.name)

    def _merge(
        self,
        instance1: Node,
        instance2: Node,
        str_instance1: str | None = None,
        str_instance2: str | None = None,
        warn_fn: Callable[[str], object] = warnings.warn,
    ):
        _Column(self.name, self.table, self.sourced)._merge(instance1, instance2, str_instance1, str_instance2, warn_fn)


class _SetAttr[T]:
    def __init__(
        self,
        table: LiteralString,
        table_column: LiteralString,
        sourced: bool = False,
        formatter: Callable[[T], T] | None = None,
    ):
        self.table = table
        self.table_column = table_column
        self.sourced = sourced
        self.formatter = formatter

    def __get__(self, instance: Node, owner: type[Node]) -> set[T]:
        return {
            v
            for (v,) in instance.conn.execute(
                f"SELECT {self.table_column} FROM {self.table} WHERE i = :i", dict(i=instance.i)
            ).fetchall()
        }

    def __set__(self, instance: Node, values: set[T] | tuple[set[int], set[T]]):
        srcs, values = values if isinstance(values, tuple) else (instance.sources, values)
        if self.formatter is not None:
            values = {self.formatter(value) for value in values}
        source_params = "?, " * (len(srcs) - 1) + "?"

        cur = instance.conn.cursor()
        if not self.sourced:
            cur.execute(f"DELETE FROM {self.table} WHERE i = :i", dict(i=instance.i))
            cur.executemany(
                f"INSERT INTO {self.table} (i, {self.table_column}) VALUES (:i, :value)",
                [dict(i=instance.i, value=value) for value in values],
            )
            return
        existing_values = {
            v
            for (v,) in cur.execute(
                f"SELECT {self.table_column} FROM {self.table + 'Source'} WHERE i = ? AND source IN ({source_params})",
                (instance.i, *srcs),
            ).fetchall()
        }
        for new_value in values - existing_values:
            cur.execute(
                f"INSERT INTO {self.table} (i, {self.table_column}) VALUES (:i, :value) "
                f"ON CONFLICT (i, {self.table_column}) DO NOTHING",
                dict(i=instance.i, value=new_value),
            )
            cur.executemany(
                f"INSERT INTO {self.table + 'Source'} (i, {self.table_column}, source) VALUES (:i, :value, :source) "
                f"ON CONFLICT (i, {self.table_column}, source) DO NOTHING",
                [dict(i=instance.i, value=new_value, source=src) for src in srcs],
            )
        for old_value in existing_values - values:
            cur.execute(
                f"DELETE FROM {self.table + 'Source'} WHERE i = ? AND {self.table_column} = ? AND source IN ({source_params})",
                (instance.i, old_value, *srcs),
            )
            if (
                cur.execute(
                    f"SELECT count(i) FROM {self.table + 'Source'} WHERE i = :i AND {self.table_column} = :value",
                    dict(i=instance.i, value=old_value),
                ).fetchone()[0]
                == 0
            ):
                cur.execute(
                    f"DELETE FROM {self.table} WHERE i = :i AND {self.table_column} = :value",
                    dict(i=instance.i, value=old_value),
                )

    def _merge(self, instance1: Node, instance2: Node, *_, **__):
        cur = instance1.conn.cursor()
        cur.execute(
            f"INSERT INTO {self.table} (i, {self.table_column}) "
            f"SELECT :i1, {self.table_column} FROM {self.table} WHERE i = :i2 "
            f"ON CONFLICT (i, {self.table_column}) DO NOTHING",
            dict(i1=instance1.i, i2=instance2.i),
        )
        if self.sourced:
            cur.execute(
                f"UPDATE OR REPLACE {self.table + 'Source'} SET i = :i1 WHERE i = :i2",
                dict(i1=instance1.i, i2=instance2.i),
            )
        cur.execute(f"DELETE FROM {self.table} WHERE i = :i2", dict(i1=instance1.i, i2=instance2.i))



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, ClassVar, Literal, NotRequired, Required, Self, TypedDict, Unpack

from gatelogue_types._util import _AircraftColumn, _Column, _FKColumn, _format_code, _format_str, _SetAttr
from gatelogue_types.node import LocatedNode, Node

if TYPE_CHECKING:
    import sqlite3
    from collections.abc import Callable, Iterator

type AirMode = Literal["helicopter", "seaplane", "warp plane", "traincarts plane"]


class Aircraft:
    """An aircraft model."""

    def __init__(self, conn: sqlite3.Connection, name: str):
        self.conn = conn
        self.name = _format_str(name)
        if self.conn.execute("SELECT 1 FROM Aircraft WHERE name = ?", (self.name,)).fetchone() is None:
            msg = f"Aircraft {self.name} does not exist"
            raise ValueError(msg)

    @property
    def manufacturer(self):
        """Manufacturing company or person of the aircraft"""
        return self.conn.execute(
            "SELECT manufacturer from Aircraft WHERE name = :name",
            dict(name=self.name),
        ).fetchone()[0]

    @manufacturer.setter
    def manufacturer(self, value: str):
        cur = self.conn.cursor()
        cur.execute(
            "UPDATE Aircraft SET manufacturer = :value WHERE name = :name",
            dict(value=value, name=self.name),
        )

    @property
    def width(self) -> int:
        """Width (Wingspan) of the aircraft"""
        return self.conn.execute(
            "SELECT width from Aircraft WHERE name = :name",
            dict(name=self.name),
        ).fetchone()[0]

    @width.setter
    def width(self, value: int):
        cur = self.conn.cursor()
        cur.execute("UPDATE Aircraft SET width = :value WHERE name = :name", dict(value=value, name=self.name))

    @property
    def height(self) -> int:
        """Height of the aircraft"""
        return self.conn.execute(
            "SELECT height from Aircraft WHERE name = :name",
            dict(name=self.name),
        ).fetchone()[0]

    @height.setter
    def height(self, value: int):
        cur = self.conn.cursor()
        cur.execute("UPDATE Aircraft SET height = :value WHERE name = :name", dict(value=value, name=self.name))

    @property
    def length(self) -> int:
        """Length of the aircraft"""
        return self.conn.execute(
            "SELECT length from Aircraft WHERE name = :name",
            dict(name=self.name),
        ).fetchone()[0]

    @length.setter
    def length(self, value: int):
        cur = self.conn.cursor()
        cur.execute("UPDATE Aircraft SET length = :value WHERE name = :name", dict(value=value, name=self.name))

    @property
    def mode(self) -> AirMode:
        """The technology the aircraft is for"""
        return self.conn.execute(
            "SELECT mode from Aircraft WHERE name = :name",
            dict(name=self.name),
        ).fetchone()[0]

    @mode.setter
    def mode(self, value: AirMode):
        cur = self.conn.cursor()
        cur.execute("UPDATE Aircraft SET length = :value WHERE name = :name", dict(value=value, name=self.name))

    @classmethod
    def create(
        cls,
        conn: sqlite3.Connection,
        *,
        name: str,
        manufacturer: str,
        width: int,
        height: int,
        length: int,
        mode: AirMode,
    ) -> Self:
        """Internal use"""
        cur = conn.cursor()
        cur.execute(
            "INSERT INTO Aircraft (name, manufacturer, width, height, length, mode) VALUES (:name, :manufacturer, :width, :height, :length, :mode)",
            dict(
                name=_format_str(name),
                manufacturer=manufacturer,
                width=width,
                height=height,
                length=length,
                mode=mode,
            ),
        )
        return cls(conn, name)


class AirAirline(Node):
    name = _Column[str]("name", "AirAirline", formatter=_format_str)
    """Name of the airline"""
    link = _Column[str | None]("link", "AirAirline", sourced=True, formatter=_format_str)
    """Link to the MRT Wiki page for the airline"""
    COLUMNS: ClassVar = (name, link)

    def __str__(self):
        return super().__str__() + f" {self.name}"

    class CreateParams(TypedDict, total=False):
        """Internal use"""

        name: Required[str]
        link: str | None

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node(conn, src, ty=cls.__name__)
        cur = conn.cursor()
        cur.execute("INSERT INTO AirAirline (i, name, link) VALUES (:i, :name, :link)", dict(i=i, **kwargs))
        cur.execute(
            "INSERT INTO AirAirlineSource (i, source, link) VALUES (:i, :source, :link_src)",
            dict(i=i, source=src, **kwargs),
        )
        return cls(conn, i)

    @property
    def flights(self) -> Iterator[AirFlight]:
        """List of all :py:class:`AirFlight` s the airline operates"""
        return (
            AirFlight(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM AirFlight WHERE airline = :i", dict(i=self.i)).fetchall()
        )

    @property
    def gates(self) -> Iterator[AirGate]:
        """List of all :py:class:`AirGate` s the airline owns or operates"""
        return (
            AirGate(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT i FROM AirGate WHERE airline = :i "
                'UNION SELECT "from" AS i FROM AirFlight WHERE airline = :i '
                'UNION SELECT "to" AS i FROM AirFlight WHERE airline = :i',
                dict(i=self.i),
            ).fetchall()
        )

    @property
    def airports(self) -> Iterator[AirAirport]:
        """List of all :py:class:`AirAirports` s the airline flies to or has gates in"""
        return (
            AirAirport(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT DISTINCT airport FROM AirGate WHERE airline = :i "
                'UNION SELECT DISTINCT airport FROM AirFlight LEFT JOIN AirGate on AirGate.i = "from" WHERE AirFlight.airline = :i '
                'UNION SELECT DISTINCT airport FROM AirFlight LEFT JOIN AirGate on AirGate.i = "to" WHERE AirFlight.airline = :i',
                dict(i=self.i),
            ).fetchall()
        )

    def equivalent_nodes(self) -> Iterator[Self]:
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM AirAirline WHERE name = ?", (self.name,)).fetchall()
        )

    def _merge(self, other: Self):
        cur = self.conn.cursor()
        cur.execute("UPDATE AirGate SET airline = :i1 WHERE airline = :i2", dict(i1=self.i, i2=other.i))
        cur.execute("UPDATE AirFlight SET airline = :i1 WHERE airline = :i2", dict(i1=self.i, i2=other.i))


class AirAirport(LocatedNode):
    @staticmethod
    def _format_airport_code(s: str) -> str:
        s = s.strip().upper()
        if len(s) == 4 and s[3] == "T":
            return s[:3]
        return s

    code = _Column[str]("code", "AirAirport", formatter=_format_airport_code)
    """Unique 3 (sometimes 4)-letter code"""
    names = _SetAttr[str]("AirAirportNames", "name", sourced=True, formatter=_format_str)
    """Name(s) of the airport"""
    link = _Column[str | None]("link", "AirAirport", sourced=True, formatter=_format_str)
    """Link to the MRT Wiki page for the airport"""
    modes = _SetAttr[AirMode]("AirAirportModes", "mode", sourced=True, formatter=_format_str)
    """Types of air vehicle or technology the airport supports"""
    COLUMNS: ClassVar = (*LocatedNode.COLUMNS, code, names, link, modes)

    def __str__(self):
        return super().__str__() + f" {self.code}"

    class CreateParams(LocatedNode.CreateParams, total=False):
        code: Required[str]
        names: set[str]
        link: str | None
        modes: set[AirMode]

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        names, modes = kwargs.get("names", set()), kwargs.get("modes", set())
        i = cls.create_node_with_location(conn, src, ty=cls.__name__, **kwargs)
        cur = conn.cursor()
        cur.execute("INSERT INTO AirAirport (i, code, link) VALUES (:i, :code, :link)", dict(i=i, **kwargs))
        cur.executemany("INSERT INTO AirAirportNames (i, name) VALUES (?, ?)", [(i, name) for name in names])
        cur.executemany("INSERT INTO AirAirportModes (i, mode) VALUES (?, ?)", [(i, mode) for mode in modes])
        cur.execute(
            "INSERT INTO AirAirportSource (i, source, link) VALUES (:i, :source, :link_src)",
            dict(i=i, source=src, **kwargs),
        )
        cur.executemany(
            "INSERT INTO AirAirportNamesSource (i, name, source) VALUES (?, ?, ?)", [(i, name, src) for name in names]
        )
        cur.executemany(
            "INSERT INTO AirAirportModesSource (i, mode, source) VALUES (?, ?, ?)", [(i, mode, src) for mode in modes]
        )
        return cls(conn, i)

    @property
    def gates(self) -> Iterator[AirGate]:
        """List of :py:class:`AirGate` s"""
        return (
            AirGate(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM AirGate WHERE airport = :i", dict(i=self.i)).fetchall()
        )

    def equivalent_nodes(self) -> Iterator[Self]:
        if (code := self.code) == "":
            return iter(())
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM AirAirport WHERE code = ?", (code,)).fetchall()
        )

    def _merge(self, other: Self):
        self.conn.execute("UPDATE AirGate SET airport = :i1 WHERE airport = :i2", dict(i1=self.i, i2=other.i))
        super()._merge(other)


class AirGate(Node):
    code = _Column[str | None]("code", "AirGate", formatter=_format_code)
    """Unique gate code. If ``None``, code is not known"""
    airport = _FKColumn(AirAirport, "airport", "AirGate")
    """The :py:class:`AirAirport`"""
    airline = _FKColumn[AirAirline | None](AirAirline, "airline", "AirGate", sourced=True)
    """The :py:class:`Airline` that owns this gate"""
    width = _Column[int | None]("width", "AirGate", sourced=True)
    """The width of this gate"""
    mode = _Column[AirMode | None]("mode", "AirGate", sourced=True, formatter=_format_str)
    """Type of air vehicle or technology the gate supports"""
    COLUMNS: ClassVar = (code, airport, airline, width, mode)

    def __str__(self):
        return super().__str__() + f" {self.airport.code} {self.code}"

    class CreateParams(TypedDict, total=False):
        """Internal use"""

        code: Required[str | None]
        airport: Required[AirAirport]
        airline: AirAirline | None
        width: int | None
        mode: AirMode | None

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node(conn, src, ty=cls.__name__)
        cur = conn.cursor()
        cur.execute(
            "INSERT INTO AirGate (i, code, airport, airline, width, mode) VALUES (:i, :code, :airport, :airline, :width, :mode)",
            dict(i=i, **kwargs),
        )
        cur.execute(
            "INSERT INTO AirGateSource (i, source, width, mode, airline) VALUES (:i, :source, :width_src, :mode_src, :airline_src)",
            dict(i=i, source=src, **kwargs),
        )
        return cls(conn, i)

    @property
    def flights_from_here(self) -> Iterator[AirFlight]:
        """List of IDs of all :py:class:`AirFlight` s that depart from this gate"""
        return (
            AirFlight(self.conn, i)
            for (i,) in self.conn.execute('SELECT i FROM AirFlight WHERE "from" = :i', dict(i=self.i)).fetchall()
        )

    @property
    def flights_to_here(self) -> Iterator[AirFlight]:
        """List of IDs of all :py:class:`AirFlight` s that arrive at this gate"""
        return (
            AirFlight(self.conn, i)
            for (i,) in self.conn.execute('SELECT i FROM AirFlight WHERE "to" = :i', dict(i=self.i)).fetchall()
        )

    def equivalent_nodes(self) -> Iterator[Self]:
        if (code := self.code) is None:
            return iter(())
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT i FROM AirGate WHERE airport = ? AND code = ?", (self.airport.i, code)
            ).fetchall()
        )

    def merge(self, other: Self, warn_fn: Callable[[str], object] = warnings.warn):
        if self.code is None and other.code is not None:
            self.code = other.code
        elif self.code is not None and other.code is None:
            other.code = self.code
        super().merge(other, warn_fn)

    def _merge(self, other: Self):
        cur = self.conn.cursor()
        cur.execute('UPDATE AirFlight SET "from" = :i1 WHERE "from" = :i2', dict(i1=self.i, i2=other.i))
        cur.execute('UPDATE AirFlight SET "to" = :i1 WHERE "to" = :i2', dict(i1=self.i, i2=other.i))


class AirFlight(Node):
    airline = _FKColumn(AirAirline, "airline", "AirFlight")
    """The :py:class:`AirAirline` the flight is operated by"""
    code = _Column[str]("code", "AirFlight", formatter=_format_code)
    """Flight code. May be duplicated if the return flight uses the same code as this flight.
    **2-letter airline prefix not included**"""
    from_ = _FKColumn(AirGate, "from", "AirFlight")
    """The :py:class:`AirGate` this flight departs from"""
    to = _FKColumn(AirGate, "to", "AirFlight")
    """The :py:class:`AirGate` this flight arrives at"""
    aircraft = _AircraftColumn("aircraft", "AirFlight", sourced=True)
    """Aircraft used on the flight"""
    COLUMNS: ClassVar = (airline, code, from_, to, aircraft)

    def __str__(self):
        return super().__str__() + f" {self.airline.name} {self.code}"

    class CreateParams(TypedDict):
        airline: AirAirline
        code: str
        from_: AirGate
        to: AirGate
        aircraft: NotRequired[str | None]

    @classmethod
    def create(
        cls,
        conn: sqlite3.Connection,
        src: int,
        **kwargs: Unpack[CreateParams],
    ) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node(conn, src, ty=cls.__name__)
        cur = conn.cursor()
        cur.execute(
            'INSERT INTO AirFlight (i, "from", "to", code, aircraft, airline) VALUES (:i, :from_, :to, :code, :aircraft, :airline)',
            dict(i=i, **kwargs),
        )
        cur.execute(
            "INSERT INTO AirFlightSource (i, source, aircraft) VALUES (:i, :source, :aircraft_src)",
            dict(i=i, source=src, **kwargs),
        )
        return cls(conn, i)

    @classmethod
    def create2(
        cls, conn: sqlite3.Connection, src: int, *, code2: str | None = None, **kwargs: Unpack[CreateParams]
    ) -> tuple[Self, Self]:
        kwargs2 = kwargs.copy()
        kwargs2["from_"], kwargs2["to"] = kwargs2["to"], kwargs2["from_"]
        kwargs2["code"] = code2 or kwargs2["code"]
        return cls.create(conn, src, **kwargs), cls.create(conn, src, **kwargs2)

    def equivalent_nodes(self) -> Iterator[Self]:
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT AirFlight.i FROM AirFlight "
                'LEFT JOIN AirGate AGFrom ON "from" = AGFrom.i '
                'LEFT JOIN AirGate AGTo ON "to" = AGTo.i '
                "WHERE AirFlight.airline = ? AND (AGFrom.airport = ? AND AGTo.airport = ?)",
                (self.airline.i, self.from_.airport.i, self.to.airport.i),
            ).fetchall()
        )

    def merge(self, other: Self, warn_fn: Callable[[str], object] = warnings.warn) -> set[int] | None:
        also_merged = set()
        if (self.from_.code is None or other.from_.code is None) and self.from_ != other.from_:
            also_merged.add(other.from_.i)
            self.from_.merge(other.from_, warn_fn=warn_fn)
        if (self.to.code is None or other.to.code is None) and self.to != other.to:
            also_merged.add(other.to.i)
            self.to.merge(other.to, warn_fn=warn_fn)
        super().merge(other, warn_fn)
        return also_merged



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Literal, Self, Unpack

from gatelogue_types._util import _Column, _format_str
from gatelogue_types.node import LocatedNode

if TYPE_CHECKING:
    import sqlite3
    from collections.abc import Iterator

type Rank = Literal["Unranked", "Councillor", "Mayor", "Senator", "Governor", "Premier", "Community"]


class Town(LocatedNode):
    name = _Column[str]("name", "Town", formatter=_format_str)
    """Name of the town"""
    rank = _Column[Rank]("rank", "Town", formatter=_format_str)
    """Rank of the town"""
    mayor = _Column[str]("mayor", "Town", formatter=_format_str)
    """Mayor of the town"""
    deputy_mayor = _Column[str | None]("deputyMayor", "Town", formatter=_format_str)
    """Deputy Mayor of the town"""
    COLUMNS: ClassVar = (*LocatedNode.COLUMNS, name, rank, mayor, deputy_mayor)

    def __str__(self):
        return super().__str__() + f" {self.name} ({self.rank})"

    class CreateParams(LocatedNode.CreateParams, total=True):
        """Internal use"""

        name: str
        rank: Rank
        mayor: str
        deputy_mayor: str | None

    @classmethod
    def create(cls, conn: sqlite3.Connection, src: int, **kwargs: Unpack[CreateParams]) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node_with_location(conn, src, ty=cls.__name__, **kwargs)
        cur = conn.cursor()
        cur.execute(
            "INSERT INTO Town (i, name, rank, mayor, deputyMayor) VALUES (:i, :name, :rank, :mayor, :deputy_mayor)",
            dict(i=i, **kwargs),
        )
        return cls(conn, i)

    def equivalent_nodes(self) -> Iterator[Self]:
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT i FROM Town WHERE name = ? and mayor = ?", (self.name, self.mayor)
            ).fetchall()
        )



"""
Python utility library for using Gatelogue data in Python projects. It will load the database for you to access via ORM or raw SQL.

Installation
------------
Run ``pip install gatelogue-types`` or ``uv add gatelogue-types``. Or add ``gatelogue-types`` to your ``requirements.txt`` or ``pyproject.toml``.

To import directly from the repository, run ``pip install git+https://github.com/mrt-map/gatelogue#subdirectory=gatelogue-types-py`` or add ``gatelogue-types @ git+https://github.com/mrt-map/gatelogue#subdirectory=gatelogue-types-py`` to your ``requirements.txt`` or ``pyproject.toml``.

You can also use ``requests``, ``niquests``, ``httpx``, ``urllib3`` or ``aiohttp`` to retrieve the data via ``gatelogue-types`` if ``[requests]``, ``[niquests]``, ``[httpx]``, ``[urllib3]`` or ``[aiohttp]`` is suffixed. Otherwise ``urllib`` is used.

Usage
-----
To retrieve the data:

.. code-block:: python

   import gatelogue_types as gt  # for convenience

   gd = gt.GD.niquests_get()  # retrieve data via niquests
   gd = gt.GD.requests_get()  # retrieve data via requests
   gd = gt.GD.httpx_get()  # retrieve data via httpx
   gd = gt.GD.urllib3_get()  # retrieve data via urllib3
   gd = gt.GD.urllib_get()  # retrieve data via urllib
   gd = await gt.GD.aiohttp_get()  # retrieve data via aiohttp

   # for all .*_get() methods, you can make it retrieve a version with sources.
   gd = gt.GD.niquests_get(sources=True)

Using the ORM does not require SQL and makes for generally clean code. However, doing this is very inefficient as each attribute access is one SQL query.

.. code-block:: python

   for airport in gd.nodes(gt.AirAirport):
       for gate in airport.gates:
           print(f"Airport {airport.code} has gate {gate.code}")

Querying the underlying SQLite database directly with ``sqlite3`` is generally more efficient and faster. It is also the only way to access the ``*Source`` tables, if you retrieved the database with those.

.. code-block:: python

   for airport_code, gate_code in gd.conn.execute(
       "SELECT A.code, G.code FROM AirGate G INNER JOIN AirAirport A ON G.airport = A.i"
   ).fetchall():
       print(f"Airport {airport.code} has gate {gate.code}")

Note that ``gatelogue-types`` *(py)* is used by ``gatelogue-aggregator``, which is why many classes have methods for modifying the database.
Usage of these methods are discouraged. These are not used in normal use 99% of the time, and they will probably error anyway.

"""

from __future__ import annotations

import contextlib
import datetime
import sqlite3
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Self,
)

from gatelogue_types.__about__ import __data_version__
from gatelogue_types.air import AirAirline, AirAirport, Aircraft, AirFlight, AirGate, AirMode
from gatelogue_types.bus import BusBerth, BusCompany, BusConnection, BusLine, BusMode, BusStop
from gatelogue_types.node import LocatedNode, Node, Proximity, SharedFacility, World
from gatelogue_types.rail import RailCompany, RailConnection, RailLine, RailMode, RailPlatform, RailStation
from gatelogue_types.sea import SeaCompany, SeaConnection, SeaDock, SeaLine, SeaMode, SeaStop
from gatelogue_types.spawn_warp import SpawnWarp, WarpType
from gatelogue_types.town import Rank, Town

if TYPE_CHECKING:
    from collections.abc import Iterator
    from os import PathLike

    # pyrefly: ignore [missing-import]
    import aiohttp

URL: str = "https://raw.githubusercontent.com/MRT-Map/gatelogue/refs/heads/dist/data.db"
URL_NO_SOURCES: str = "https://raw.githubusercontent.com/MRT-Map/gatelogue/refs/heads/dist/data-ns.db"


class GD:
    """
    Main class that contains an :py:class:`sqlite3.Connection`
    """

    conn: sqlite3.Connection
    """Connection to the underlying SQL database"""

    def __init__(self, database: str | bytes | PathLike[str] | PathLike[bytes] = ":memory:"):
        sqlite3.threadsafety = 3
        self.conn = sqlite3.connect(database, check_same_thread=False, autocommit=True)
        self.conn.execute("PRAGMA foreign_keys = true")

    @classmethod
    def from_bytes(cls, data: bytes) -> Self:
        self = cls()
        self.conn.deserialize(data)
        return self

    @classmethod
    def from_bytes_readonly(cls, data: bytes) -> Self:
        self = cls()
        self.conn.deserialize(data)
        self.conn.execute("PRAGMA query_only = true")
        return self

    @classmethod
    def niquests_get(cls, sources: bool = False, *args, **kwargs):
        # pyrefly: ignore [missing-import]
        import niquests

        return cls.from_bytes_readonly(niquests.get(URL if sources else URL_NO_SOURCES, *args, **kwargs).content)

    @classmethod
    def requests_get(cls, sources: bool = False, *args, **kwargs):
        # pyrefly: ignore [missing-import,missing-source-for-stubs]
        import requests

        return cls.from_bytes_readonly(requests.get(URL if sources else URL_NO_SOURCES, *args, **kwargs).content)

    @classmethod
    def httpx_get(cls, sources: bool = False, *args, **kwargs):
        # pyrefly: ignore [missing-import]
        import httpx

        return cls.from_bytes_readonly(httpx.get(URL if sources else URL_NO_SOURCES, *args, **kwargs).content)

    @classmethod
    def urllib3_get(cls, sources: bool = False, *args, **kwargs):
        # pyrefly: ignore [missing-import]
        import urllib3

        return cls.from_bytes_readonly(urllib3.request("GET", URL if sources else URL_NO_SOURCES, *args, **kwargs).data)

    @classmethod
    def urllib_get(cls, sources: bool = False, *args, **kwargs):
        import urllib.request

        with urllib.request.urlopen(URL if sources else URL_NO_SOURCES, *args, **kwargs) as response:  # noqa: S310
            return cls.from_bytes_readonly(response.read())

    @classmethod
    async def aiohttp_get(cls, sources: bool = False, session: aiohttp.ClientSession | None = None) -> Self:
        # pyrefly: ignore [missing-import]
        import aiohttp

        session = aiohttp.ClientSession() if session is None else contextlib.nullcontext(session)
        async with (
            session as session,
            session.get(URL if sources else URL_NO_SOURCES) as response,
        ):  # pyrefly: ignore [missing-attribute]
            return cls.from_bytes_readonly(await response.read())

    @property
    def timestamp(self) -> str:
        """Time that the aggregation of the data was done"""
        return self.conn.execute("SELECT timestamp FROM Metadata").fetchone()[0]

    @property
    def version(self) -> int:
        """Version number of the database format"""
        return self.conn.execute("SELECT version FROM Metadata").fetchone()[0]

    @property
    def has_sources(self) -> bool:
        """Whether the database has sources"""
        return self.conn.execute("SELECT has_sources FROM Metadata").fetchone()[0]

    @classmethod
    def create(cls, sources: list[str], database=":memory:", has_sources: bool = True) -> Self:
        """Internal Use"""
        self = cls(database=database)
        cur = self.conn.cursor()
        cur.executescript((Path(__file__).parent / "create.sql").read_text())
        cur.execute(
            "INSERT INTO Metadata (version, timestamp, has_sources) VALUES (:version, :timestamp, :has_sources)",
            dict(version=__data_version__, timestamp=datetime.datetime.now().isoformat(), has_sources=has_sources),
        )

        if has_sources:
            cur.executemany("INSERT INTO Source (priority, name) VALUES (?, ?)", list(enumerate(sources)))
        else:
            self.drop_sources(cur)
        self.conn.commit()
        return self

    def drop_sources(self, cur: sqlite3.Cursor | None = None):
        """Drop all ``*Source`` tables"""
        cur = cur or self.conn.cursor()
        cur.executescript((Path(__file__).parent / "drop_sources.sql").read_text())

    def get_node[T: Node = Node](self, i: int, ty: type[T] | None = None) -> T:
        """Get a single node

        If ``ty`` is ``None``, :py:class:`LocatedNode` or :py:class:`Node`, find the node type automatically.
        Set ``ty`` to a specific node type only if you can guarantee that the node of ID ``i`` is of said node type."""
        if ty is None or ty is Node:
            return Node.auto_type(self.conn, i)
        if ty is LocatedNode:
            return LocatedNode.auto_type(self.conn, i)
        return ty(self.conn, i)

    def nodes[T: Node = Node](self, ty: type[T] | None = None) -> Iterator[T]:
        """Get all nodes, optionally of a specific type."""
        if ty is None or ty is Node:
            return (
                Node.STR2TYPE[ty](self.conn, i) for i, ty in self.conn.execute("SELECT i, type FROM Node").fetchall()
            )
        if ty is LocatedNode:
            return (
                Node.STR2TYPE[ty](self.conn, i)
                for i, ty in self.conn.execute(
                    "SELECT NodeLocation.i, type FROM NodeLocation LEFT JOIN Node on Node.i = NodeLocation.i"
                ).fetchall()
            )
        return (
            ty(self.conn, i)
            for (i,) in self.conn.execute("SELECT i FROM Node WHERE type = :type", dict(type=ty.__name__))
        )

    def __iter__(self) -> Iterator[Node]:
        return self.nodes()

    def __len__(self):
        return self.conn.execute("SELECT count(rowid) FROM Node").fetchone()[0]



from __future__ import annotations

from typing import TYPE_CHECKING, ClassVar, Literal, Self, Unpack

from gatelogue_types._util import _Column, _format_str
from gatelogue_types.node import LocatedNode

if TYPE_CHECKING:
    import sqlite3
    from collections.abc import Iterator

type WarpType = Literal["premier", "terminus", "traincarts", "portal", "misc"]


class SpawnWarp(LocatedNode):
    name = _Column[str]("name", "SpawnWarp", formatter=_format_str)
    """Name of the spawn warp"""
    warp_type = _Column[WarpType]("warpType", "SpawnWarp", formatter=_format_str)
    """The type of the spawn warp"""
    COLUMNS: ClassVar = (*LocatedNode.COLUMNS, name, warp_type)

    def __str__(self):
        return super().__str__() + f" {self.name} ({self.warp_type})"

    class CreateParams(LocatedNode.CreateParams, total=True):
        """Internal use"""

        name: str
        warp_type: str

    @classmethod
    def create(
        cls,
        conn: sqlite3.Connection,
        src: int,
        **kwargs: Unpack[CreateParams],
    ) -> Self:
        """Internal use"""
        kwargs = cls.format_create_kwargs(**kwargs)
        i = cls.create_node_with_location(conn, src, ty=cls.__name__, **kwargs)
        cur = conn.cursor()
        cur.execute(
            "INSERT INTO SpawnWarp (i, name, warpType) VALUES (:i, :name, :warp_type)",
            dict(i=i, **kwargs),
        )
        return cls(conn, i)

    def equivalent_nodes(self) -> Iterator[Self]:
        return (
            type(self)(self.conn, i)
            for (i,) in self.conn.execute(
                "SELECT i FROM SpawnWarp WHERE name = ? and warpType = ?", (self.name, self.warp_type)
            ).fetchall()
        )



from __future__ import annotations

import warnings
from typing import TYPE_CHECKING, ClassVar, Literal, Self, TypedDict, Unpack

from gatelogue_types._util import _AircraftColumn, _Column, _CoordinatesColumn, _FKColumn, _format_str, _SetAttr

if TYPE_CHECKING:
    import sqlite3
    from collections.abc import Callable, Iterable, Iterator


class Node:
    """Base class of all nodes"""

    STR2TYPE: ClassVar[dict] = {}
    """Internal use"""

    def __init_subclass__(cls, **kwargs):
        cls.STR2TYPE[cls.__name__] = cls

    def __init__(self, conn: sqlite3.Connection, i: int):
        self.conn = conn
        assert isinstance(i, int), i
        self.i = i
        """The ID of the node"""

    @classmethod
    def auto_type(cls, conn: sqlite3.Connection, i: int) -> Node:
        """Automatically find and get the :py:class:`Node` subclass from the index ``i``"""
        (ty,) = conn.execute("SELECT type FROM Node WHERE i = :i", dict(i=i)).fetchone()
        return cls.STR2TYPE[ty](conn, i)

    type = _Column[str]("type", "Node")
    """The type of the node"""
    sources = _SetAttr[int]("NodeSource", "source")
    """All sources that prove the node's existence. Does not exist in no-source databases."""
    COLUMNS: ClassVar[tuple[_Column | _FKColumn | _SetAttr | _CoordinatesColumn, ...]] = ()

    @property
    def source(self) -> int:
        """Returns a single source.
        :exception AssertionError: if this node has more than one source, or if querying a non-source database."""
        sources = self.sources
        assert len(sources) == 1, "Expected only one source"
        return next(iter(sources))

    @source.setter
    def source(self, value: int):
        self.sources = {value}

    def __str__(self):
        return type(self).__name__ + f"({self.i})"

    @classmethod
    def create_node(cls, conn: sqlite3.Connection, src: int, *, ty: str) -> int:
        """Internal use"""
        cur = conn.cursor()
        cur.execute("INSERT INTO Node ( type ) VALUES ( :type )", dict(type=ty))
        (i,) = cur.execute("SELECT i FROM Node WHERE ROWID = last_insert_rowid()").fetchone()
        cur.execute("INSERT INTO NodeSource ( i, source ) VALUES ( :i, :source )", dict(i=i, source=src))
        return i

    def __eq__(self, other: Self) -> bool:
        return self.i == other.i

    def __hash__(self):
        return hash(self.i)

    def equivalent_nodes(self) -> Iterator[Self]:
        """Internal use"""
        raise NotImplementedError

    def _merge(self, other: Self):
        pass

    def merge(self, other: Self, warn_fn: Callable[[str], object] = warnings.warn):
        """Internal use"""
        if self == other:
            warn_fn(f"{self} tried to merge with itself")
            return
        self_str = str(self)
        other_str = str(other)
        for attr in self.COLUMNS:
            attr._merge(self, other, self_str, other_str, warn_fn)

        self._merge(other)
        other.delete()

    def delete(self):
        """Internal use"""
        cur = self.conn.cursor()
        cur.execute(f"DELETE FROM {type(self).__name__}Source WHERE i = :i", dict(i=self.i))
        cur.execute(f"DELETE FROM {type(self).__name__} WHERE i = :i", dict(i=self.i))
        cur.execute("DELETE FROM NodeSource WHERE i = :i", dict(i=self.i))
        cur.execute("DELETE FROM Node WHERE i = :i", dict(i=self.i))

    @classmethod
    def format_create_kwargs(cls, **kwargs) -> dict:
        """Internal use"""
        for attr in cls.COLUMNS:
            key = (attr.table_column + "s" if isinstance(attr, _SetAttr) else attr.name).strip('"')
            if key == "from":
                key += "_"
            if key not in kwargs:
                kwargs[key] = None
            if isinstance(attr, _Column):
                if attr.formatter is not None:
                    kwargs[key] = attr.formatter(kwargs[key])
                if attr.sourced:
                    kwargs[key + "_src"] = kwargs[key] is not None
            elif isinstance(attr, _SetAttr):
                if attr.formatter is not None:
                    kwargs[key] = {attr.formatter(value) for value in kwargs[key]} if kwargs[key] is not None else set()
                if attr.sourced:
                    kwargs[key + "_src"] = kwargs[key] is not None and len(kwargs[key]) > 0
            elif isinstance(attr, _FKColumn):
                kwargs[key] = kwargs[key].i if kwargs[key] is not None else None
                if attr.sourced:
                    kwargs[key + "_src"] = kwargs[key] is not None
            elif isinstance(attr, _CoordinatesColumn):
                kwargs["coordinates_src"] = kwargs["coordinates"] is not None
                kwargs["x"] = int(kwargs["coordinates"][0]) if kwargs["coordinates"] is not None else None
                kwargs["y"] = int(kwargs["coordinates"][1]) if kwargs["coordinates"] is not None else None
            elif isinstance(attr, _AircraftColumn):
                if kwargs[key] is not None:
                    kwargs[key] = kwargs[key].strip()
                if attr.sourced:
                    kwargs[key + "_src"] = kwargs[key] is not None
        return kwargs


type World = Literal["New", "Old", "Space"]


class LocatedNode(Node):
    """Base class of all nodes with a location (world + coordinates)"""

    world = _Column[World | None]("world", "NodeLocation", sourced=True, formatter=_format_str)
    """The world the node is in"""
    coordinates = _CoordinatesColumn()
    """The coordinates of the node"""
    COLUMNS: ClassVar = (world, coordinates)

    @classmethod
    def auto_type(cls, conn: sqlite3.Connection, i: int) -> LocatedNode:
        (ty,) = conn.execute(
            "SELECT type FROM NodeLocation LEFT JOIN Node on Node.i = NodeLocation.i WHERE NodeLocation.i = :i",
            dict(i=i),
        ).fetchone()
        return cls.STR2TYPE[ty](conn, i)

    class CreateParams(TypedDict, total=False):
        """Internal use"""

        world: World | None
        coordinates: tuple[int, int] | None

    @classmethod
    def create_node_with_location(
        cls, conn: sqlite3.Connection, src: int, *, ty: str, **kwargs: Unpack[CreateParams]
    ) -> int:
        """Internal use"""
        i = cls.create_node(conn, src, ty=ty)
        cur = conn.cursor()
        cur.execute(
            "INSERT INTO NodeLocation (i, world, x, y) VALUES (:i, :world, :x, :y)",
            dict(i=i, **kwargs),
        )
        cur.execute(
            "INSERT INTO NodeLocationSource (i, source, world, coordinates) VALUES (:i, :source, :world_src, :coordinates_src)",
            dict(i=i, source=src, **kwargs),
        )
        return i

    @property
    def _nodes_in_proximity(self) -> Iterator[int]:
        return (
            j if self.i == i else i
            for i, j in self.conn.execute(
                "SELECT node1, node2 FROM Proximity WHERE node1 = :i OR node2 = :i", dict(i=self.i)
            ).fetchall()
        )

    @property
    def nodes_in_proximity(self) -> Iterator[tuple[LocatedNode, Proximity]]:
        """
        References all nodes that are near (within walking distance of) this object.

        :return: Pairs of nodes in proximity as well as proximity data (:py:class:`Proximity`).
        """
        return (
            (o_node := LocatedNode.auto_type(self.conn, o), Proximity(self.conn, self, o_node))
            for o in self._nodes_in_proximity
        )

    @property
    def _shared_facilities(self) -> Iterator[int]:
        return (
            j if self.i == i else i
            for i, j in self.conn.execute(
                "SELECT node1, node2 FROM SharedFacility WHERE node1 = :i OR node2 = :i", dict(i=self.i)
            ).fetchall()
        )

    @property
    def shared_facilities(self) -> Iterable[LocatedNode]:
        """References all nodes that this object shares the same facility with (same building, station, hub etc)"""
        return (LocatedNode.auto_type(self.conn, o) for o in self._shared_facilities)

    def _merge(self, other: Self):
        cur = self.conn.cursor()
        cur.execute("UPDATE OR FAIL SharedFacility SET node1 = :i1 WHERE node1 = :i2", dict(i1=self.i, i2=other.i))
        cur.execute("UPDATE OR FAIL SharedFacility SET node2 = :i1 WHERE node2 = :i2", dict(i1=self.i, i2=other.i))
        cur.execute(
            "INSERT INTO Proximity SELECT min(node2, :i1), max(node2, :i1), distance, explicit FROM Proximity "
            "WHERE node1 = :i2 ON CONFLICT (node1, node2) DO NOTHING",
            dict(i1=self.i, i2=other.i),
        )
        cur.execute(
            "INSERT INTO Proximity SELECT min(node1, :i1), max(node1, :i1), distance, explicit FROM Proximity "
            "WHERE node2 = :i2 ON CONFLICT (node1, node2) DO NOTHING",
            dict(i1=self.i, i2=other.i),
        )
        cur.execute(
            "UPDATE OR FAIL ProximitySource SET (node1, node2) = (min(node2, :i1), max(node2, :i1)) WHERE node1 = :i2",
            dict(i1=self.i, i2=other.i),
        )
        cur.execute(
            "UPDATE OR FAIL ProximitySource SET (node1, node2) = (min(node1, :i1), max(node1, :i1)) WHERE node2 = :i2",
            dict(i1=self.i, i2=other.i),
        )
        cur.execute("DELETE FROM Proximity WHERE node1 == :i2 OR node2 == :i2", dict(i1=self.i, i2=other.i))

    def delete(self):
        cur = self.conn.cursor()
        cur.execute(f"DELETE FROM {type(self).__name__}Source WHERE i = :i", dict(i=self.i))
        cur.execute(f"DELETE FROM {type(self).__name__} WHERE i = :i", dict(i=self.i))
        cur.execute("DELETE FROM NodeLocationSource WHERE i = :i", dict(i=self.i))
        cur.execute("DELETE FROM NodeLocation WHERE i = :i", dict(i=self.i))
        cur.execute("DELETE FROM NodeSource WHERE i = :i", dict(i=self.i))
        cur.execute("DELETE FROM Node WHERE i = :i", dict(i=self.i))


class Proximity:
    """Proximity data between two :py:class:`LocatedNode` s"""

    def __init__(self, conn: sqlite3.Connection, node1: LocatedNode, node2: LocatedNode):
        self.conn = conn
        if node1.i > node2.i:
            node1, node2 = node2, node1
        self.node1 = node1.i
        self.node2 = node2.i

    @property
    def distance(self) -> float:
        """Distance between the two objects in blocks"""
        return self.conn.execute(
            "SELECT distance from Proximity WHERE node1 = :node1 AND node2 = :node2",
            dict(node1=self.node1, node2=self.node2),
        ).fetchone()[0]

    @distance.setter
    def distance(self, value: float):
        cur = self.conn.cursor()
        cur.execute(
            "UPDATE Proximity set distance = :value WHERE node1 = :node1 AND node2 = :node2",
            dict(value=value, node1=self.node1, node2=self.node2),
        )

    @property
    def explicit(self) -> bool:
        """Whether this relation is explicitly recognised by the company/ies of the stations. Used mostly for local services"""
        return self.conn.execute(
            "SELECT explicit from Proximity WHERE node1 = :node1 AND node2 = :node2",
            dict(node1=self.node1, node2=self.node2),
        ).fetchone()[0]

    @explicit.setter
    def explicit(self, value: bool):
        cur = self.conn.cursor()
        cur.execute(
            "UPDATE Proximity set explicit = :value WHERE node1 = :node1 AND node2 = :node2",
            dict(value=value, node1=self.node1, node2=self.node2),
        )

    @classmethod
    def create(
        cls,
        conn: sqlite3.Connection,
        srcs: Iterable[int],
        *,
        node1: LocatedNode,
        node2: LocatedNode,
        distance: float,
        explicit: bool = False,
    ) -> Self:
        """Internal use"""
        if node1.i > node2.i:
            node1, node2 = node2, node1

        cur = conn.cursor()
        cur.execute(
            "INSERT INTO Proximity (node1, node2, distance, explicit) VALUES (:node1, :node2, :distance, :explicit) ON CONFLICT DO NOTHING",
            dict(node1=node1.i, node2=node2.i, distance=distance, explicit=explicit),
        )
        cur.executemany(
            "INSERT INTO ProximitySource ( node1, node2, source ) VALUES ( :node1, :node2, :source ) ON CONFLICT DO NOTHING",
            [dict(node1=node1.i, node2=node2.i, source=src) for src in srcs],
        )
        return cls(conn, node1, node2)


class SharedFacility:
    """Shared facility data between two :py:class:`LocatedNode` s. Currently has no attributes."""

    def __init__(self, conn: sqlite3.Connection, node1: LocatedNode, node2: LocatedNode):
        self.conn = conn
        if node1.i > node2.i:
            node1, node2 = node2, node1
        self.node1 = node1.i
        self.node2 = node2.i

    @classmethod
    def create(cls, conn: sqlite3.Connection, *, node1: LocatedNode, node2: LocatedNode) -> Self:
        """Internal use"""
        if node1.i > node2.i:
            node1, node2 = node2, node1

        cur = conn.cursor()
        cur.execute(
            "INSERT INTO SharedFacility (node1, node2) VALUES (:node1, :node2) ON CONFLICT DO NOTHING ",
            dict(node1=node1.i, node2=node2.i),
        )
        return cls(conn, node1, node2)
